#define VERSION			"0.017"
#define DEFAULTPORT		6667
#define COMMANDCHAR		'/'
#define HISTORY_SIZE	40

/* colors / display control */
#define OFF             0
#define FLASH			5
#define INVERSE			7
#define BLACK			30
#define RED				31
#define GREEN			32
#define YELLOW			33
#define BLUE			34
#define MAGENTA			35
#define CYAN			36
#define WHITE			37
#define INTENSIFIED		37

#define BLACK_BG		40
#define RED_BG			41
#define GREEN_BG		42
#define YELLOW_BG		43
#define BLUE_BG			44
#define MAGENTA_BG		45
#define CYAN_BG			46
#define WHITE_BG_I		47
#define NORMAL_BG		48

/* for ignore */
#define NONE			0	/* default */
#define CTCP			1
#define PUBLIC			2	/* channel */
#define INVITES			4
#define NOTICES			8
#define MESSAGES		16
#define WALLOPS			32
#define NOTES			64 
#define ALL				65535

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <strings.h>
#include <string.h>
#include <termios.h>
#include <errno.h>
#include <sys/types.h>
#include <pwd.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <sys/un.h>
#include <setjmp.h>
     